/*
  menu-generator - CLI application generating menus for various WMs
  Copyright (C) 2012 Jaromir Capik <tavvva@email.cz>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"

#define WM_ICEWM    1
#define WM_FLUXBOX  2
#define WM_BLACKBOX 3

/* returns true if string A begins with substring B */
#define begins_with(a,b) (strncmp(a,b,strlen(b))==0)

/* returns true if string A ends with substring B */
#define ends_with(a,b) (strlen(a)<strlen(b))?0:(strncmp(a+strlen(a)-strlen(b),b,strlen(b))==0)

/* search in A till B char is reached (A set to B position) */
#define find_char(a,b) while((a[0]!='\0') && (a[0]!=b)) a++;

/* search in A till B char is reached (A set to position of B successor) */
#define find_char_successor(a,b) find_char(a,b); if (a[0]==b) a++;

/* strip CR at the end of string */
#define strip_cr(a) if((strlen(a)>0) && (a[strlen(a)-1]=='\n')) a[strlen(a)-1]='\0';


/* filename string */
typedef char t_filename[_POSIX_PATH_MAX + 1];

/* group string */
typedef char t_group[ENTRY_GROUP_MAX_LEN + 1];

/* name string */
typedef char t_name[ENTRY_NAME_MAX_LEN + 1];

/* category string */
typedef char t_category[DESKTOP_FILE_LINE_MAX_LEN + 1];


/* menu entry */
typedef struct {
  t_name name;
  t_filename icon;
  t_filename exec;
  t_group *group;
} t_menu_entry;

/* group entry */
typedef struct {
  t_category category;
  t_group group;
  unsigned char localization_strength;  /* 0=none, 1=short, 2=full */
  t_filename icon;
} t_group_entry;

/* array of menu entries */
t_menu_entry menu[ENTRIES_MAX];
int menu_entries=0;

/* array of all supported groups */
t_group_entry groups[GROUP_ENTRIES_MAX];
int group_entries=0;

/* subset of really used groups and their icons */
t_group *used_groups[GROUP_ENTRIES_MAX];
t_filename *used_group_icons[GROUP_ENTRIES_MAX];
int used_group_entries=0;

/* list of generated groups */
t_group already_generated_groups[GROUP_ENTRIES_MAX];
int generated_group_entries=0;

/* strings holding the current locale */
char locale[LOCALE_MAX_LEN + 1];
char short_locale[LOCALE_MAX_LEN + 1];

/* string holding the user group definion filename */
char user_group_definitions[_POSIX_PATH_MAX + 1];

/* strings holding variable names for localized Name entries */
char name_locale[LOCALE_MAX_LEN + 1];
char name_short_locale[LOCALE_MAX_LEN + 1];

/* WM session - icewm, ... */
char *wm_session;
int wm_index;

/* reads the group definition file */
/* Return: 0-failure, 1-success */
int read_group_definitions(char *group_definitions_filename)
{
  FILE *f;
  static char line[DESKTOP_FILE_LINE_MAX_LEN+1];
  char *tmp,*tmp2,*tmp3,*value;
  signed char localization_strength;

  f=fopen(group_definitions_filename,"r");
  if (f == NULL) {
    return 0; /* error message is returned in the calling function */
  }

  group_entries=-1;
  while (fgets(line, sizeof(line), f) != NULL) {

    strip_cr(line);

    value=line; find_char_successor(value,'=');
    tmp=line;  find_char(tmp,'='); tmp[0]='\0';   /* line holds the whole variable name + possible braces */
    tmp2=line; find_char(tmp2,']'); tmp2[0]='\0'; /* line holds the whole variable name + possible left brace&locale */
    tmp3=line; find_char_successor(tmp3,'[');     /* tmp3 holds possible locale */
    tmp2=line; find_char(tmp2,'['); tmp2[0]='\0'; /* line holds the whole variable name ONLY */

    /* check, how much localized the entry is */ /* 0=none, 1=short, 2=full */
    if (strlen(tmp3) == 0) {
      localization_strength = 0;
    } else if (strcmp(tmp3,short_locale) == 0) {
      localization_strength = 1;
    } else if (strcmp(tmp3,locale) == 0) {
      localization_strength = 2;
    } else if (strcmp(tmp3,":icon") == 0) { /* special entry for icon */
      localization_strength = -1;
    } else continue; /* unwanted language - skip the entry */

    if ((group_entries < 0) || (strcmp(groups[group_entries].category, line)!=0)) {
      group_entries++; /* very first entry or group name changed -> increase index */
      if (group_entries>=GROUP_ENTRIES_MAX) {
        fprintf(stderr,"Error: Maximum number of supported groups reached!\n");
        fclose(f);
        return 0;
      }
      strcpy(groups[group_entries].icon,"");
      groups[group_entries].localization_strength=0; /* initial localization strength */
      strncpy(groups[group_entries].category, line, DESKTOP_FILE_LINE_MAX_LEN);
      groups[group_entries].category[DESKTOP_FILE_LINE_MAX_LEN] = '\0';
    }

    if (groups[group_entries].localization_strength <= localization_strength) {
      strncpy(groups[group_entries].group, value, ENTRY_GROUP_MAX_LEN);
      groups[group_entries].group[ENTRY_GROUP_MAX_LEN] = '\0';
      groups[group_entries].localization_strength = localization_strength;
    }

    if (localization_strength == -1) { /* special value for icon */
      strncpy(groups[group_entries].icon, value, _POSIX_PATH_MAX);
    }

  }

  fclose(f);

  return 1;
}



t_group *find_group(char *category) {
  int i;
  char *tmp,*last,*tmp2;

  for (i=0;i<=group_entries;i++) { /* for each group entry */

    /* reset position */
    tmp2=category;

    /* check match with any of the fields in the category entry */
    last = &tmp2[strlen(tmp2)];

    while (tmp2!=last) {
      tmp=tmp2;
      find_char(tmp,';');
      tmp[0]='\0';
      if (strcmp(tmp2,groups[i].category)==0) {  /* group found */
        return &groups[i].group;
      }
      tmp2=tmp; if (tmp2<last) tmp2++[0]=';'; /* fix erased semicolon and continue to next field */
    }

  }

  /* unknown group */
  return (t_group *)"";
}


/* Add entry in the menu array */
/* Return: 1 when successful, 0 when failed */
int add_entry(t_filename desktop_file_name)
{
  FILE *f;
  int localization_strength;
  static char line[DESKTOP_FILE_LINE_MAX_LEN+1];
  char *tmp;

  /* menu array full */
  if (menu_entries >= ENTRIES_MAX) {
    fprintf(stderr,"Error : menu array full!\n");
    return 0;
  }

  f=fopen(desktop_file_name,"r");
  if (f == NULL) {
    fprintf(stderr,"Error : can't open .desktop file!\n");
    return 0;
  }

  menu[menu_entries].group = (t_group *)""; /* init group to empty */
  localization_strength=0;
  while (fgets(line, sizeof(line), f) != NULL) {

    strip_cr(line);

    /* process generic Name entry ... lowest priority */
    if ((localization_strength==0) && (begins_with(line,"Name="))) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      strncpy(menu[menu_entries].name,tmp,ENTRY_NAME_MAX_LEN);
      menu[menu_entries].name[ENTRY_NAME_MAX_LEN]='\0';
    }

    /* process localized Name entry (short locale definition) ... middle priority */
    if ((localization_strength<2) && (begins_with(line,name_short_locale))) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      strncpy(menu[menu_entries].name,tmp,ENTRY_NAME_MAX_LEN);
      menu[menu_entries].name[ENTRY_NAME_MAX_LEN]='\0';
      localization_strength=1;
    }

    /* process localized Name entry (full locale definition) ... highest priority */
    if (begins_with(line,name_locale)) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      strncpy(menu[menu_entries].name,tmp,ENTRY_NAME_MAX_LEN);
      menu[menu_entries].name[ENTRY_NAME_MAX_LEN]='\0';
      localization_strength=2;
    }

    /* process Icon entry */
    if (begins_with(line,"Icon=")) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      strncpy(menu[menu_entries].icon,tmp,_POSIX_PATH_MAX);
      menu[menu_entries].icon[_POSIX_PATH_MAX]='\0';
    }

    /* process Exec entry */
    if (begins_with(line,"Exec=")) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      strncpy(menu[menu_entries].exec,tmp,_POSIX_PATH_MAX);
      menu[menu_entries].exec[_POSIX_PATH_MAX]='\0';
    }

    /* process Category(Group) entry */
    if (begins_with(line,"Categories=")) {
      /* cut the value */
      tmp=line;
      find_char_successor(tmp,'=');
      menu[menu_entries].group = find_group(tmp);
    }
  }
  fclose(f);

  menu_entries++;

  return 1;
}


/* add group to list of used groups (if not exists) */
void add_used_group_entry(t_group *group, t_filename *icon)
{
  int i;

  for (i=0;i<used_group_entries;i++) {
    if (used_groups[i]==group) return; /* group already in the list */
  }

  used_group_icons[used_group_entries]=icon;
  used_groups[used_group_entries++]=group; /* index shouldn't overrun */
                                           /* since used groups are subset of all groups */
                                           /* and share the same limit definition */
}

/* add group to list of generated groups (if not exists) */
int add_already_generated(char *group)
{
  int i;

  for (i=0;i<generated_group_entries;i++) {
    if (strcmp(already_generated_groups[i],group)==0) return 1; /* group already in the list */
  }

  strcpy(already_generated_groups[generated_group_entries++], group); /* index shouldn't overrun */
                                                   /* since used groups are subset of all groups */
                                                          /* and share the same limit definition */
  return 0;
}

/* calculate the submenu depth */
int prefix_level(char *prefix)
{
  int level;
  if (strlen(prefix)==0) return 0;
  level=1;
  if (begins_with(prefix,"Menu-header") || begins_with(prefix,"Menu-footer")) level--;
  while (prefix[0]!='\0') {
    if (prefix++[0]=='|') level++;
  }
  return level;
}

/* generate output to stdout - begin of the menu/submenu block */
void generate_block_begin(char *prefix)
{
  int i,pl;
  char *tail;
  char *group_icon="";

  /* search for the group icon */
  for (i=0;i<used_group_entries;i++) {
    if (strcmp((char *)used_groups[i],prefix) == 0) {
      group_icon=(char *)used_group_icons[i];
      break;
    }
  }

  pl=prefix_level(prefix);
  tail=prefix;
  if (begins_with(prefix,"Menu-header") || begins_with(prefix,"Menu-footer")) find_char_successor(tail,'|');
  for (i=1;i<pl;i++) {
    find_char_successor(tail,'|');
    printf("  ");
  }

  switch (wm_index) {
    case WM_ICEWM:
      if (strlen(group_icon)==0) group_icon="folder";
      if (pl>0) printf("menu \"%s\" \"%s\" {\n",tail,group_icon);
      break;
    case WM_FLUXBOX:
      if (strlen(group_icon)==0) group_icon="folder";
      if (pl==0) printf("[begin] (Menu)\n[encoding] {UTF-8}\n");
      if (pl>0) printf("[submenu] (%s) <%s/%s%s> {\n",tail,ICON_FILES_LOCATION,group_icon,FLUXBOX_ICON_SUFFIX);
      break;
    case WM_BLACKBOX:
      if (pl==0) printf("[begin] (Menu)\n");
      if (pl>0) printf("[submenu] (%s) {\n",tail);
      break;
  }

}

/* generate output to stdout - end of the menu/submenu block */
void generate_block_end(char *prefix)
{
  int i,pl;
  pl=prefix_level(prefix);
  for (i=1;i<pl;i++) printf("  ");

  switch (wm_index) {
    case WM_ICEWM:
      if (pl>0) printf("}\n");
      if (pl==1) printf("\n");
      break;
    case WM_FLUXBOX:
      if (pl==0) {
        printf("[separator]\n");
        printf("[submenu] (Fluxbox)\n");
        printf("  [config] (Configure)\n");
        printf("  [submenu] (System Styles) {Choose a style...}\n");
        printf("    [stylesdir] (/usr/share/fluxbox/styles)\n");
        printf("  [end]\n");
        printf("  [submenu] (User Styles) {Choose a style...}\n");
        printf("    [stylesdir] (~/.fluxbox/styles)\n");
        printf("  [end]\n");
        printf("  [workspaces] (Workspace List)\n");
        /*printf("  [submenu] (Tools)\n");*/
        printf("  [exec] (Window name) {xprop WM_CLASS|cut -d \" -f 2|xmessage -file - -center}\n");
        /*printf("  [exec] (Regen Menu) {../util/fluxbox-generate_menu }\n");*/
        /*printf("[end]\n");*/
        printf("  [commanddialog] (Fluxbox Command)\n");
        printf("  [reconfig] (Reload config)\n");
        printf("  [restart] (Restart)\n");
        printf("  [exec] (About) {(fluxbox -v; fluxbox -info | sed 1d) 2> /dev/null | xmessage -file - -center}\n");
        printf("  [separator]\n");
        printf("  [exit] (Exit)\n");
        printf("[end]\n");
        printf("[endencoding]\n[end]\n");
      }
      if (pl>0) printf("[end]\n");
      if (pl==1) printf("\n");
      break;
    case WM_BLACKBOX:
      if (pl==0) {
        printf("[exec] (——————————) {echo}\n");
        printf("[submenu] (Blackbox)\n");
        printf("  [config] (Configure)\n");
        printf("  [submenu] (System Styles) {Choose a style...}\n");
        printf("    [stylesdir] (/usr/share/blackbox/styles)\n");
        printf("  [end]\n");
        printf("  [submenu] (User Styles) {Choose a style...}\n");
        printf("    [stylesdir] (~/.blackbox/styles)\n");
        printf("  [end]\n");
        printf("  [workspaces] (Workspace List)\n");
        printf("  [exec] (Window name) {xprop WM_CLASS|cut -d \" -f 2|xmessage -file - -center}\n");
        printf("  [reconfig] (Reload config)\n");
        printf("  [restart] (Restart)\n");
        printf("  [exec] (——————————) {echo}\n");
        printf("  [exit] (Exit)\n");
        printf("[end]\n");
        printf("[end]\n");
      }
      if (pl>0) printf("[end]\n");
      if (pl==1) printf("\n");
      break;
  }
}

/* generate output to stdout - separator */
void generate_separator(char *prefix,int initial_newline)
{
  int i,pl;
  pl=prefix_level(prefix);
  for (i=0;i<pl;i++) printf("  ");
  if (initial_newline) printf("\n");

  switch (wm_index) {
    case WM_ICEWM:
      printf("separator\n\n");
      break;
    case WM_FLUXBOX:
      printf("[separator]\n\n");
      break;
    case WM_BLACKBOX:
      printf("[exec] (——————————) {echo}\n\n");
      break;
  }
}

/* generate output to stdout - menu entry */
void generate_entry(t_menu_entry *entry)
{
  int i,pl;
  pl=prefix_level((char *)entry->group);
  for (i=0;i<pl;i++) printf("  ");

  switch (wm_index) {
    case WM_ICEWM:
      printf("prog \"%s\" \"%s\" %s\n",entry->name,entry->icon,entry->exec);
      break;
    case WM_FLUXBOX:
      if (strlen(entry->icon)==0) {
        printf("[exec] (%s) {%s}\n",entry->name,entry->exec);
      } else {
        printf("[exec] (%s) {%s} <%s/%s%s>\n",entry->name,entry->exec,ICON_FILES_LOCATION,entry->icon,FLUXBOX_ICON_SUFFIX);
      }
      break;
    case WM_BLACKBOX:
      printf("[exec] (%s) {%s}\n",entry->name,entry->exec);
      break;
  }
}

/* recursively generate submenu matching the prefix */
void generate_submenu(char *prefix)
{
  int i;
  char *tmp;
  t_group tmp2;

  /* add in the list of already generated submenus */
  if (add_already_generated(prefix)) return;

  /* create new submenu */
  /* if the prefix is empty, then we're in the top level menu :) That could help us for WMs without top level block definition */
  generate_block_begin(prefix);

  /* add all entries from the header group in the top level menu */
  if (strlen(prefix)==0) {
    for (i=0;i<menu_entries;i++) {
      if (strcmp((char *)menu[i].group,"Menu-header")==0) generate_entry(&menu[i]); /* entry in the header */
      if (begins_with((char *)menu[i].group,"Menu-header|")) { /* submenu in the header */

        /* generate new prefix - just one more level */
        strcpy(tmp2,(char *)menu[i].group);
        tmp=tmp2;
        tmp+=strlen("Menu-header|");
        find_char(tmp,'|'); tmp[0]='\0';

        /* generate submenu in the header */
        generate_submenu(tmp2);
      }
    }
    generate_separator(prefix,1);
  }

  /* find submenus */
  for (i=0;i<used_group_entries;i++) {

    if ( (strlen((char *)(used_groups[i])) > strlen(prefix)) &&
         begins_with((char *)used_groups[i],prefix) && 
         ( (strlen(prefix)==0) || (((char *)used_groups[i])[strlen(prefix)]=='|') || (((char *)used_groups[i])[strlen(prefix)]=='\0') ) ) {

      /* generate new prefix - just one more level */
      strcpy(tmp2,(char *)used_groups[i]);
      tmp=tmp2;
      tmp+=strlen(prefix)+1;
      find_char(tmp,'|');tmp[0]='\0';

      /* generate submenu except the header/footer group */
      if ( (strcmp(tmp2,"Menu-header")!=0) && (strcmp(tmp2,"Menu-footer")!=0) )  generate_submenu(tmp2);
    }

  }

  /* add all entries in this level (matching the prefix) */
  for (i=0;i<menu_entries;i++) {
    if (strcmp((char *)menu[i].group,prefix)==0) generate_entry(&menu[i]);
  }

  /* add all entries from the footer group in the top level menu */
  if (strlen(prefix)==0) {
    generate_separator(prefix,1);
    for (i=0;i<menu_entries;i++) {
      if (strcmp((char *)menu[i].group,"Menu-footer")==0) generate_entry(&menu[i]);
      if (begins_with((char *)menu[i].group,"Menu-footer|")) { /* submenu in the footer */

        /* generate new prefix - just one more level */
        strcpy(tmp2,(char *)menu[i].group);
        tmp=tmp2;
        tmp+=strlen("Menu-footer|");
        find_char(tmp,'|'); tmp[0]='\0';

        /* generate submenu in the footer */
        generate_submenu(tmp2);
      }
    }
  }

  /* close the block */
  generate_block_end(prefix);

}

/* scan the specified directory containing .desktop diles and fill the menu array with entries */
int add_desktop_files_from_location(char *location)
{
  struct dirent **filelist = {0};
  int fcount = -1;
  int i = 0;

  struct stat fstat;
  t_filename full_name;

  size_t path_len;

  /* Read the directory */
  fcount = scandir(location, &filelist, 0, alphasort);

  if(fcount < 0) {
    fprintf(stderr,"Error: Directory scan failed!\n");
    return 0;
  }

  path_len=strlen(location);

  for (i=0; i<fcount; i++) { /* For each directory entry... */

    /* check if the full name length is in the length limit */
    if ((path_len + strlen(filelist[i]->d_name) + 1) > _POSIX_PATH_MAX) { /* if not, discard the entry :( */
      free(filelist[i]);
      continue;
    }

    /* ignore special directories */
    if ((strcmp(filelist[i]->d_name, ".") == 0) || (strcmp(filelist[i]->d_name, "..") == 0)) {
      free(filelist[i]);
      continue;
    }

    strcpy(full_name, location);
    if (full_name[path_len - 1] != '/') strcat(full_name, "/");
    strcat(full_name, filelist[i]->d_name); 
    free(filelist[i]);

    /* process only regular files */
    if (stat(full_name, &fstat) < 0) continue;
    if (S_ISREG(fstat.st_mode)) {
      /* Select only .desktop files */
      if (ends_with(full_name,".desktop")) add_entry(full_name);
    }

  }

  /* Finalize resources. */
  free(filelist);

  return 1;
}


int main(int argc, char **argv)
{
  char *tmp;
  int cnt,cnt2;

  /* Read the target WM */
  if (argc==2) wm_session=argv[1];
  else {
    fprintf(stderr,"\nError: Wrong number of arguments!\n\n");
    fprintf(stderr,"Usage: menu-generator <target_window_manager>\n\n");
    fprintf(stderr,"Supported WMs:\n");
    fprintf(stderr,"  icewm\n");
/*    fprintf(stderr,"  fluxbox\n");*/
    fprintf(stderr,"\n");
    return EXIT_FAILURE;
  }

  /* Check if the WM is supported */
  if (strcmp(wm_session,"icewm")==0) wm_index=WM_ICEWM;
  else if (strcmp(wm_session,"fluxbox")==0) wm_index=WM_FLUXBOX;
  else if (strcmp(wm_session,"blackbox")==0) wm_index=WM_BLACKBOX;
  else {
    fprintf(stderr,"Error: Unsupported WM!\n");
    return EXIT_FAILURE;
  }

  /* Create user group definitions filename */
  tmp = getenv("HOME");
  if (tmp == NULL) {
    fprintf(stderr,"Error: HOME variable not set! Can't continue!\n");
    return EXIT_FAILURE;
  }
  strncpy(user_group_definitions,tmp,_POSIX_PATH_MAX);
  user_group_definitions[_POSIX_PATH_MAX] = '\0';
  strncat(user_group_definitions,USER_GROUP_DEFINITIONS,_POSIX_PATH_MAX);
  user_group_definitions[_POSIX_PATH_MAX] = '\0';

  /* Detect current locale from LANG */
  tmp = getenv("LANG");
  if (tmp == NULL) tmp="";

  strncpy(locale, tmp, LOCALE_MAX_LEN);
  locale[LOCALE_MAX_LEN] = '\0';

  /* Cut the suffix (.UTF-8 etc) */
  tmp=locale;
  find_char(tmp,'.');
  tmp[0]='\0';

  /* Cut the region specific part (_whatever) */
  strcpy(short_locale,locale);
  tmp=short_locale;
  find_char(tmp,'_');
  tmp[0]='\0';

  /* Creating variable names for localized Name entries */
  snprintf(name_locale, LOCALE_MAX_LEN, "Name[%s]=", locale);
  name_locale[LOCALE_MAX_LEN]=0;
  snprintf(name_short_locale, LOCALE_MAX_LEN, "Name[%s]=", short_locale);
  name_short_locale[LOCALE_MAX_LEN]=0;

  /* Read group definitions */
  if (!read_group_definitions(user_group_definitions)) {
    if (!read_group_definitions(GROUP_DEFINITIONS)) {
      fprintf(stderr,"Error: Can't read menu-generator-groups definition file!\n");
      return EXIT_FAILURE;
    }
  }

  /* try to read .desktop files */
  if (!add_desktop_files_from_location(DESKTOP_FILES_LOCATION)) return EXIT_FAILURE;

  /* generate list of used groups - sorted according to the list of supported groups */
  for (cnt=0;cnt<=group_entries;cnt++) {
    for (cnt2=0;cnt2<menu_entries;cnt2++) {
      if (&groups[cnt].group==menu[cnt2].group) add_used_group_entry(&groups[cnt].group,&groups[cnt].icon);
    }
  }

  /* recursive menu generation */
  generate_submenu("");

  return EXIT_SUCCESS;
}
